<?php

/*
Plugin Name: Fotobook
Plugin URI: http://www.aaronharp.com/dev/wp-fotobook/
Description: This plugin allows you to import Facebook photo galleries directly into WordPress.
Author: Aaron Harp
Author URI: http://www.aaronharp.com/
Version: 1.2
*/

$fb_album_table = $table_prefix.'fb_albums';
$fb_photo_table = $table_prefix.'fb_photos';
$fb_posts_table = $table_prefix.'posts';
$fb_plugin_path = ABSPATH.'/wp-content/plugins/fotobook/';
$fb_plugin_url  = get_option('siteurl').'/wp-content/plugins/fotobook/';
$fb_styles_file = $fb_plugin_path.'/fotobook-styles.css';
$fb_website     = 'http://www.aaronharp.com/dev/wp-fotobook/';

function fb_add_manage_page() {
	add_management_page('Fotobook Management', 'Fotobook', 8, 'fotobook/manage-fotobook.php');
}

function fb_add_options_page() {
  add_options_page('Fotobook Options', 'Fotobook', 8, 'fotobook/options-fotobook.php');
}

function fb_initialize() {
	global $fb_photo_table, $fb_album_table;
	
	// add default options
  add_option('fb_albums_page', 0);
  add_option('fb_number_rows', 5);
  add_option('fb_number_cols', 4);
  add_option('fb_album_cmts', 1);
  add_option('fb_thumb_size', 130);
		
	function table_exists($table_name) {
	  global $wpdb;
	  foreach ($wpdb->get_col("SHOW TABLES",0) as $table ) {
	    if ($table == $table_name) return true;
    }
	  return false;
	}
	
	function create_table($table_name, $query) {
	  global $wpdb;
	  $result = $wpdb->query($query);
	  return table_exists($table_name);
	}
    
  $photo_table_query = "CREATE TABLE $fb_photo_table (
                        id int(11) unsigned NOT NULL auto_increment,
                        album_id int(11) NOT NULL default '0',
                        title varchar(255) NOT NULL default '',
                        full varchar(255) NOT NULL default '',
                        thumb varchar(255) NOT NULL default '',
                        UNIQUE KEY id (id)
                        ) TYPE = MyISAM";

  $album_table_query = "CREATE TABLE $fb_album_table (
                          id int(11) unsigned NOT NULL auto_increment,
                          page_id int(11) NOT NULL,
                          url varchar(255) NOT NULL default '',
                          name varchar(255) NOT NULL default '',
                          description text NOT NULL default '',
                          location varchar(255) NOT NULL default '',
                          photo_count int(11) NOT NULL default 0,
                          ordinal int(11) NOT NULL default 0,
                          timecached datetime NOT NULL,
                          UNIQUE KEY  id (id)
                        ) TYPE = MyISAM";  

	if(!table_exists($fb_photo_table)) {
	  create_table($fb_photo_table, $photo_table_query);
	}
	
	if(!table_exists($fb_album_table)) {
	  create_table($fb_photo_table, $album_table_query);
	}

}

function fb_add_album($url) {
  global $wpdb, $fb_album_table, $fb_photo_table;
  
  if(!strstr($url,'.facebook.com')) {
    return "That is not a valid Facebook album.";
  }
  
  // check if this album is already imported
  $row = $wpdb->get_row("SELECT * FROM $fb_album_table WHERE url = '$url'", ARRAY_A);
  if($row != '') return "You have already imported that album.  Use the re-cache button to reimport it.";
  
  // get the album
  $album = fb_parse_album($url);
  if(!$album) return "There was an error importing the album.  Please check your link.";
  $photo_count = sizeof($album['photos']);
  
  // add slashes to user input
  function slash(&$string) { addslashes($string); }
  slash($album['name']);
  slash($album['location']);
  slash($album['description']);
  
  // create the WP page
  $page_id = fb_add_page($album['name']);
  
  // move all albums up one ordinal
  if(fb_get_lowest_ordinal() == 0) {
    $wpdb->query("UPDATE $fb_album_table SET ordinal = ordinal + 1");
  }
  
  // add to database
  $album_query = sprintf("INSERT INTO %s SET page_id=%s, url='%s', name='%s', location='%s', description='%s', photo_count=%s, ordinal=%s, timecached=NOW()",
                        $fb_album_table, $page_id, $album['url'], $album['name'], $album['location'], $album['description'], $photo_count, 0);  
  $wpdb->query($album_query);
  
  // get id of album that was just added
  $album_id = mysql_insert_id();
  
  // now add photos to db
  foreach($album['photos'] as $photo) {
    $photo_query = sprintf("INSERT INTO %s SET album_id=%s, title='%s', full='%s', thumb='%s'",
                           $fb_photo_table, $album_id, $photo['title'], $photo['full'], $photo['thumb']);
    $wpdb->query($photo_query);
  }
  
  return true;
}

function fb_change_order($album_id, $inc = 1) {
  global $wpdb, $fb_album_table;
    
  if($inc != 1 && $inc != -1)
    return false;
  
  // select current and new ordinal
  $op   = $inc == 1 ? '>='  : '<='; 
  $ord  = $inc == 1 ? 'ASC' : 'DESC';
  $curr_ord = $wpdb->get_var("SELECT ordinal FROM $fb_album_table WHERE id = $album_id LIMIT 1");
  $rows = $wpdb->get_results("SELECT id, ordinal FROM $fb_album_table WHERE ordinal $op $curr_ord ORDER BY ordinal $ord LIMIT 2", ARRAY_A);
  $curr_ord = $rows[0]['ordinal']; $curr_id = $rows[0]['id'];
  $new_ord  = $rows[1]['ordinal']; $new_id  = $rows[1]['id'];
  
  // check if this is the first or last one
  if($inc == 1  && $rows[0]['ordinal'] == fb_get_highest_ordinal())
    return true;
  if($inc == -1 && $rows[0]['ordinal'] == fb_get_lowest_ordinal())
    return true;
    
  // otherwise update with new ordinals
  $wpdb->query("UPDATE $fb_album_table SET ordinal = $new_ord  WHERE id = $curr_id");
  $wpdb->query("UPDATE $fb_album_table SET ordinal = $curr_ord WHERE id = $new_id");  
}

function fb_get_lowest_ordinal() {
  global $wpdb, $fb_album_table;
  return $wpdb->get_var("SELECT ordinal FROM $fb_album_table ORDER BY ordinal ASC LIMIT 1");  
}

function fb_get_highest_ordinal() {
  global $wpdb, $fb_album_table;  
  return $wpdb->get_var("SELECT ordinal FROM $fb_album_table ORDER BY ordinal DESC LIMIT 1");
}

function fb_recache_album($id) {
  global $wpdb, $fb_album_table, $fb_photo_table, $fb_posts_table;

  // get the url and page_id
  $url = $wpdb->get_var("SELECT url FROM $fb_album_table WHERE id = $id");
  $page_id = $wpdb->get_var("SELECT page_id FROM $fb_album_table WHERE id = $id");
  
  // get rid of all the photos in this album
  $wpdb->query("DELETE FROM $fb_photo_table WHERE album_id = $id");    
  
  // get the album
  $album = fb_parse_album($url);
  $photo_count = sizeof($album['photos']);
  if(!$album) return "There was an error importing the album. The link may be invalid now.  Try deleting the album and then re-importing it.";
  
  // add slashes to user input
  function slash(&$string) { addslashes($string); }
  slash($album['name']);
  slash($album['location']);
  slash($album['description']);
  
  // check if wp page still exists
  $page_row = $wpdb->get_row('SELECT * FROM '.$fb_posts_table.' WHERE ID = '.$page_id);
  if(isset($page_row->ID)) {
    fb_update_page($page_id, $album['name']);
  } else {
    $page_id = fb_add_page($album['name']);
  }

  // update the database
  $album_query = sprintf("UPDATE %s SET page_id=%s, name='%s', location='%s', description='%s', photo_count=%s, timecached=NOW() WHERE id=$id",
                        $fb_album_table, $page_id, $album['name'], $album['location'], $album['description'], $photo_count);  
  $wpdb->query($album_query);
    
  // now add photos to db
  foreach($album['photos'] as $photo) {
    $photo_query = sprintf("INSERT INTO %s SET album_id=%s, title='%s', full='%s', thumb='%s'",
                           $fb_photo_table, $id, $photo['title'], $photo['full'], $photo['thumb']);
    $wpdb->query($photo_query);
  }
  
  return true;
}

function fb_delete_album($id) {
  global $fb_photo_table, $fb_album_table, $wpdb;
  
  // delete page
  $page_id = $wpdb->get_var("SELECT page_id FROM $fb_album_table WHERE id = $id");
  fb_delete_page($page_id);
  
  $wpdb->query("DELETE FROM $fb_album_table WHERE id = $id");
  $wpdb->query("DELETE FROM $fb_photo_table WHERE album_id = $id");
}

function fb_get_url($url) {
  // first try to use cURL
  if(function_exists('curl_init')) {
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_CONNECTTIMEOUT,10);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
    curl_setopt($ch, CURLOPT_USERAGENT, "MSIE");
    $data = curl_exec($ch);
    if(curl_error($ch)) {
      curl_close($ch);
      return false;
    } else {
      return $data;
    }
  }
  // otherwise change php's user agent temporarily and use file_get_contents
  else {
    ini_set('user_agent','MSIE');
    $data = file_get_contents($url); 
    if(strstr($date,'Incompatible')) {
      return false;
    } else {
      return $data;
    } 
  }
}

function fb_parse_album($url) {
  // initialize the array for the album
  $album = array();
  $album['url'] = $url;
  
  // get facebook sub domain
  preg_match('#http://([^\.]*).facebook.com#', $url, $matches);
  $sub = $matches[1];
    
  // get the html of the album
  $album_html = array();
  if(!$album_html[0] = fb_get_url($url)) return false;
  
  // get album name
  preg_match('#<h1>.*? Photos &ndash; ([^<]*)</h1>#', $album_html[0], $matches);
  $album['name'] = $matches[1];
  
  // get album description
  preg_match('#<div class="description">(.*?)</div>#', $album_html[0], $matches);
  $album['description'] = $matches[1];
  
  // get album location
  preg_match('#<div class="location">Location: (.*?)</div>#', $album_html[0], $matches);
  $album['location'] = $matches[1];
  
  // if it couldn't find the name then it didn't get the right page
  // so return false
  if($album['name'] == '') {
    return false;
  }
  
  // now get html of all the other albums
  preg_match_all('#<a href="/album.php([^"]*?)">[0-9]+</a>#is', $album_html[0], $out);
  for($i = 0; $i < (sizeof($out[1]) / 2); $i++) {
    $album_html[] = fb_get_url('http://'.$sub.'.facebook.com/album.php'.$out[1][$i]);
  }
  
  // get all photos
  $album['photos'] = array();
  foreach($album_html as $html) {
    preg_match_all('#<td>[^<]*?<a href="[^"]*?">[^<]*?<img id="[^"]*?" src="([^"]*?)" title="([^"]*?)"#is', $html, $matches);  
    for($i = 0; $i < sizeof($matches[1]); $i++) {
      $album['photos'][] = array('thumb' => $matches[1][$i], 'title' => $matches[2][$i]);
    }
  }
  
  // now get all the image urls
  for($i = 0; $i < sizeof($album['photos']); $i++) {
    $album['photos'][$i]['full'] = preg_replace('#(s)([^/s]*?\.jpg)#', 'n$2', $album['photos'][$i]['thumb']);
  }
  
  return $album;
  
}

function fb_add_page($name) {
  global $wpdb, $fb_posts_table;
  
  $title = addslashes($name);
  $name = sanitize_title($name);
  
  $fb_albums_page = get_option('fb_albums_page');
  $fb_album_cmts = get_option('fb_album_cmts');
  if($fb_album_cmts) $comment_status = 'open';
  else $comment_status = 'closed';
  
  $page = array(
    'post_author'=> 1,
    'post_status'=>'static',
    'post_content'=>'',
    'post_title'=>$title,
    'post_name'=>$name,
    'comment_status'=>$comment_status,
    'post_parent'=>$fb_albums_page
  );
  
  return wp_insert_post($page);
}

function fb_delete_page($id) {
  wp_delete_post($id);
}

function fb_update_page($id, $name) {
  global $wpdb, $fb_posts_table;
  
  $title = addslashes($name);
  $name = sanitize_title($name);
    
  $fb_albums_page = get_option('fb_albums_page');
  $fb_album_cmts = get_option('fb_album_cmts');
  if($fb_albums_cmts) 
    $comment_status = 'open';
  else 
    $comment_status = 'closed';
  
  $array = array('post_author'=>1,
                 'post_status'=>'static',
                 'post_title'=>$title,
                 'post_name'=>$name,
                 'post_category'=>0,
                 'comment_status'=>$comment_status,
                 'post_parent'=>$fb_albums_page,
                 'ID'=>$id);
  
  wp_update_post($array);
  
  /*$wpdb->query("UPDATE $fb_posts_table SET 
    post_author = 1,
    post_status = 'static',
    post_title = '$title',
    post_name = '$name',
    post_category = 0,
    comment_status = '$comment_status',
    post_modified = NOW(),
    post_modified_gmt = NOW(),
    post_parent = $fb_albums_page
    WHERE ID = $id
  ");*/
  
  return true;
}

function fb_options_update_albums_page($new_id) {
  global $wpdb, $fb_posts_table;
    
  $old_id = get_option('fb_albums_page');
  if($old_id == $new_id) {
    return true;
  }
  
  $albums = fb_get_album();
  if(sizeof($albums) > 0) {
    foreach($albums as $album) {
      $wpdb->query("UPDATE $fb_posts_table SET post_parent = $new_id WHERE ID = ".$album['page_id']);
    }
  }
  
  update_option('fb_albums_page', $new_id);  
}

function fb_options_toggle_comments($status = true) {
  global $wpdb, $fb_posts_table;
  
  if($status) $status = 'open';
  else $status = 'closed';
  
  $fb_albums_page = get_option('fb_albums_page');
  
  $wpdb->query("UPDATE $fb_posts_table SET comment_status='$status' WHERE post_parent = $fb_albums_page");
}

function fb_albums_page_is_set() {
  $album_page = get_option('fb_albums_page');
  $post = get_post($album_page);
  
  if($album_page == 0 || !is_object($post)) {
    return false;
  } else {
    return true;
  }
}

function fb_get_album($album_id = 0) {
  global $wpdb, $fb_album_table;

  $query = "SELECT * FROM $fb_album_table ";
  
  if($album_id) {
    $query .= "WHERE id = $album_id";
    $array = $wpdb->get_results($query, ARRAY_A);
    return $array[0];
  }
  
  $query .= "ORDER BY ordinal ASC";
  
  return $wpdb->get_results($query, ARRAY_A);
}

function fb_get_album_id($page_id) {
  global $wpdb, $fb_album_table;
  return $wpdb->get_var("SELECT id FROM $fb_album_table WHERE page_id = $page_id");
}

function fb_get_random_photo($album_id = 0, $count = 1) {
  global $wpdb, $fb_photo_table;
  $query = "SELECT * FROM $fb_photo_table ";
  if($album_id != 0) $query .= "WHERE album_id = $album_id ";
  $query .= "ORDER BY rand() LIMIT $count";
  
  return $wpdb->get_row($query, ARRAY_A);
}

function fb_get_photos($album_id = 0) {
  global $wpdb, $fb_photo_table;

  $query = "SELECT * FROM $fb_photo_table ";
  if($album_id != 0) $query .= "WHERE album_id = $album_id ";
  
  $query .= "ORDER BY id ASC";
  
  return $wpdb->get_results($query, ARRAY_A);
}

function fb_album_display($content) {    
  global $fb_plugin_url, $wpdb, $fb_posts_table, $fb_website;
    
  // get variables to check if this is part of fotobook
  $post = $GLOBALS['post'];
  $post_id          = $post->ID;
  $post_parent      = $post->post_parent;
  $albums_page_id   = get_option('fb_albums_page');
  $albums_page_link = get_permalink($albums_page_id);
  $post_link        = get_permalink($post_id);  
  
  if($post_id != $albums_page_id && $post_parent != $albums_page_id) {
    return $content;
  }
  
  // display main photo gallery with album thumbs
  if($post_id == $albums_page_id) {
    remove_filter('the_content','wpautop');
        
    // buffer the output
    ob_start();
    
    // ADD THE STYLES
    echo '<style type="text/css">';
    $styles   = file_get_contents($fb_plugin_url.'fotobook-styles.css');
    $styles   = str_replace("\n", "", $styles); // remove new lines so that WP doesn't
    echo str_replace("\r", "", $styles); // add new p's and br's
    echo '</style>';  
    
    $albums = fb_get_album();
    $album_count = sizeof($albums);
    if($album_count == 0) $message = 'There are no photo albums.';
    if($album_count == 1) $message = 'There is 1 photo album.';
    if($album_count >  1) $message = 'There are '.$album_count.' photo albums.';
    ?> 
    <div class="fotobook-subheader">
      <span class="main"><?php echo $message ?></span>&nbsp;
    </div>
    
    <table id="fotobook-main">
    <?php
    if(sizeof($albums) > 0):
    foreach($albums as $album):
    $random_photo = fb_get_random_photo($album['id']);
    $album_link = get_permalink($album['page_id']);
    $thumb = $random_photo['thumb'];
    ?>
      <tr>
        <th>
          <a href="<?php echo $album_link?>"><img src="<?php echo $thumb ?>" border='0'></a>
        </th>
        <td>
          <a href='<?php echo  $album_link ?>'><?php echo $album['name']; ?></a><br />
          <?php if($album['description'] != ''): echo $album['description'] ?><br /><?php endif; ?>
          <small><?php echo $album['photo_count'] ?> photos</small>
        </td>
      </tr>
    <?php 
    endforeach; 
    endif;
    ?>
    </table>
    <div class="fotobook-subheader fotobook-subheader-bottom">
      <small>Powered by <a href="<?php echo $fb_website; ?>" target="_blank">Fotobook</a></small>
    </div>
       
  <?php
    // now capture the buffer and add it to $content
    $content .= ob_get_clean();
    
  } // end if that checks if this is a photo album
  
  // display individual albums
  if($post_parent == $albums_page_id && $post_parent != 0) {
    // turn off content filter so that <p> and <br> tags aren't added
    remove_filter('the_content','wpautop');

    // buffer the output
    ob_start();

    $album_id = fb_get_album_id($post_id);
    $album = fb_get_album($album_id);
    $photos = fb_get_photos($album_id);
    $photo_count = sizeof($photos);
    
    $thumb_size = get_option('fb_thumb_size');
    $number_cols = get_option('fb_number_cols');
    $number_rows = get_option('fb_number_rows');
    if($number_rows == 0) $number_rows = ceil($photo_count / $number_cols);
    $photos_per_page = $number_cols * $number_rows;
    
    $page_count = ceil($photo_count / $photos_per_page);
    $curr_page = isset($_GET['album_p']) && $_GET['album_p'] < $page_count ? $_GET['album_p'] : 0;
    $first_photo = $curr_page * $photos_per_page;
    $last_photo = $first_photo + $photos_per_page - 1;
    $last_photo = $last_photo >= $photo_count ? $photo_count - 1 : $last_photo;
    $rows_curr_page = ceil(($last_photo + 1 - $first_photo) / $number_cols);
    
    // generate pagination
    $prev_link = $curr_page != 0 ? $curr_page - 1 : null;
    if($prevlink)
      $prev_link = $post_link.(strstr($post_link, '?') ? '&album_p='.($prev_link) : '?album_p='.($prev_link));
    $next_link = $curr_page + 1 < $page_count ? $curr_page + 1 : null;
    if($next_link)
      $next_link = $post_link.(strstr($post_link, '?') ? '&album_p='.($next_link) : '?album_p='.($next_link));        
    $pagination = '';
    for($i = 0; $i < $page_count; $i++) {
      if($i == $curr_page) 
        $pagination .= '<b>'.($i+1).'</b>';
      else {
        $link = $post_link.(strstr($post_link, '?') ? '&album_p='.$i : '?album_p='.$i);
        $pagination .= "<a href='$link'>".($i+1)."</a>";
      }
    }
    
    // album info
    $description = $album['description'];
    $location = $album['location'];
    
    // ADD THE STYLES
    echo '<style type="text/css">';
    $styles   = file_get_contents($fb_plugin_url.'fotobook-styles.css');
    $styles   = str_replace("\n", "", $styles); // remove new lines so that WP doesn't
    echo str_replace("\r", "", $styles); // add new p's and br's
    echo '</style>';  
        
    // add hidden links for all images before so that next and previous
    // buttons in lightbox will display these images as well
    $hidden_top = ''; $hidden_bottom = '';
    for($i = 0; $i < $first_photo; $i++) {
      $hidden_top .= "<a href=\"{$photos[$i]['full']}\" rel=\"lightbox[fotobook]\" title=\"{$photos[$i]['title']}\"></a>";
    }
    for($i = $last_photo+1; $i < $photo_count; $i++) {
      $hidden_bottom .= "<a href=\"{$photos[$i]['full']}\" rel=\"lightbox[fotobook]\" title=\"{$photos[$i]['title']}\"></a>";
    }

    ?>
    <p style="display: none"><?php echo $hidden_top ?></p>
    
    <?php // DISPLAY PAGINATION AND RETURN LINK ?>
    <div class="fotobook-subheader">
      <span class='main'>Photos <?php echo ($first_photo+1)." - ".($last_photo+1) ?> out of <?php echo $photo_count ?> | <a href='<?php echo $albums_page_link ?>'>Back to Albums</a></span>
      <div class='pagination'>
        <?php if($prev_link): ?><a href='<?php $prev_link ?>'>Prev</a><?php endif; ?>
        <?php echo $pagination ?>
        <?php if($next_link): ?><a href='<?php echo $next_link ?>'>Next</a><?php endif; ?>
      </div>
    </div>

    <?php // BUILD THE PHOTO TABLE ?>
    <table id="fotobook-album">
      <?php
      // row loop
      $curr = $first_photo;
      for($i = 0; $i < $rows_curr_page; $i++): ?>
      <tr>
      <?php
        // column loop
        for($j = 0; $j < $number_cols; $j++):
        if($curr <= $last_photo): ?>
        <td>
          <a href='<?php echo $photos[$curr]['full'] ?>' rel='lightbox[fotobook]' title='<?php echo $photos[$curr]['title'] ?>' target='_blank'>
            <img src='<?php echo $photos[$curr]['thumb'] ?>' style='max-width: <?php echo $thumb_size ?>px; max-height: <?php echo $thumb_size ?>px; width: expression(this.width > <?php echo $thumb_size ?> ? <?php echo $thumb_size ?>: true); height: expression(this.height > <?php echo $thumb_size ?> ? <?php echo $thumb_size ?>: true);'>
          </a>
        </td>
        <?php else: ?>
        <td>&nbsp;</td>
        <?php 
        endif; 
        $curr++; 
      endfor; ?>
      </tr>      
    <?php endfor; ?>
    </table>
    
    <div class="fotobook-subheader fotobook-subheader-bottom">
      <span class='main'>Photos <?php echo ($first_photo+1)." - ".($last_photo+1) ?> out of <?php echo $photo_count ?> | <a href='<?php echo $albums_page_link ?>'>Back to Albums</a></span>
      <div class='pagination'>
        <?php if($prev_link): ?><a href='<?php $prev_link ?>'>Prev</a><?php endif; ?>
        <?php echo $pagination ?>
        <?php if($next_link): ?><a href='<?php echo $next_link ?>'>Next</a><?php endif; ?>
      </div>
    </div>
    
    <?php // DISPLAY THE ALBUM INFO  ?>
    <table id="fotobook-info">
    <?php if($description): ?>
      <tr>
        <th>Description:</th>
        <td><?php echo $description ?></td>
      </tr>
    <?php endif; ?>
    <?php if($location): ?>
      <tr>
        <th>Location:</th>
        <td><?php echo $location ?></td>
      </tr>
    <?php endif; ?>
    </table>

    <?php  // add hidden links for the rest ?>
    <p style="display: none"><?php echo $hidden_bottom ?></p>
  <?php 
    $content .= ob_get_clean();
  } // end if that checks if this is a gallery
  
  return $content;
}

if(!function_exists('file_put_contents')) {
  function file_put_contents($n,$d) {
    $f=@fopen($n,"w");
    if (!$f) {
     return false;
    } else {
     fwrite($f,$d);
     fclose($f);
     return true;
    }
  }
}

// if plugin has just been activated, then initialize Fotobook
if(isset($_GET['activate'])) {
  fb_initialize();
}

add_action('admin_head', 'fb_add_manage_page');
add_action('admin_head', 'fb_add_options_page');
add_filter('the_content', 'fb_album_display');

?>
